--[[
	Know-It-All
	Written by pceric
	http://www.wowinterface.com/downloads/info9179-Know-It-All.html
	
	Big thanks to Limited, Naturfreund, the Ace 3 team and other contributors.
	
	TODO: Interface for adding/removing spells
		  Calculate spell stacks
		  Ability to split bars into separate groups
]]
KnowItAll = LibStub("AceAddon-3.0"):NewAddon("KnowItAll", "AceEvent-3.0", "AceTimer-3.0", "AceConsole-3.0", "LibBars-1.0")
local L = LibStub("AceLocale-3.0"):GetLocale("KnowItAll", true)
local media = LibStub("LibSharedMedia-3.0")

KnowItAll.title = L["Know-It-All"]
KnowItAll.version = GetAddOnMetadata("KnowItAll", "Version")

-- Shared media
media:Register("sound", "UT2K4: Spree", [[Interface\AddOns\KnowItAll\media\killingspree.mp3]])
media:Register("sound", "UT2K4: Rampage", [[Interface\AddOns\KnowItAll\media\rampage.mp3]])
media:Register("sound", "UT2K4: Dominating", [[Interface\AddOns\KnowItAll\media\dominating.mp3]])
media:Register("sound", "UT2K4: Unstoppable", [[Interface\AddOns\KnowItAll\media\unstoppable.mp3]])
media:Register("sound", "UT2K4: Godlike", [[Interface\AddOns\KnowItAll\media\godlike.mp3]])
media:Register("sound", "UT2K4: Wicked", [[Interface\AddOns\KnowItAll\media\wickedsick.mp3]])
media:Register("statusbar", "BantoBarReverse", [[Interface\AddOns\KnowItAll\media\BantoBarReverse.tga]])

-- List of combat events we want to parse
local COMBAT_EVENTS = {
	["PARTY_KILL"] = true,
	["SPELL_AURA_APPLIED"] = true,
	["SPELL_AURA_APPLIED_DOSE"] = true, -- triggers on the 2nd or more times a stacking buff/debuff is applied
	["SPELL_AURA_REFRESH"] = true,  -- only has dest
	["SPELL_AURA_REMOVED"] = true,
	["SPELL_AURA_REMOVED_DOSE"] = true,
	["SPELL_CAST_SUCCESS"] = true,  -- needed for combo points
--	["SPELL_DAMAGE"] = true,  -- needed for immolate or other dmg+dot spells (not needed as of 3.0.2)
--	["SPELL_ENERGIZE"] = true,
--	["SPELL_HEAL"] = true,  -- needed for regrowth or other heal+hot spells (not needed as of 3.0.2)
	["SPELL_SUMMON"] = true,
	["UNIT_DESTROYED"] = true,
	["UNIT_DIED"] = true,
}

local defaults = {
	profile = {
		locked = false,
		flash = true,
		sounds = true,
		debug = false,
		maxBars = 10,
		cbTexture = "BantoBarReverse",
		bLength = 200,
		cbScale = 0.9,
		cbSpacing = 0,
		orientation = 1,
		debuffBars = true,
		debuffHideI = false,
		debuffHideB = false,
		debuffColor = { 1, 0, 0 },
		globalDebuffs = true,
		gainBars = true,
		gainHideI = false,
		gainHideB = false,
		gainColor = { 0, 1, 0 },
		globalGains = true,
		ccBars = true,
		ccHideI = false,
		ccHideB = false,
		ccColor = { 0, 0, 1 },
		invulnBars = true,
		invulnHideI = false,
		invulnHideB = false,
		invulnColor = { 1, 1, 0 },
		summonBars = true,
		summonHideI = false,
		summonHideB = false,
		summonColor = { 0.9, 0, 1 },
		cbX = 470,
		cbY = 275,
	},
}

local db
local myGUID -- your GUID
local myClass -- your class
local ksCounter = 0  -- number of players we have killed without dying
local barTable = {} -- our bar DB
local drTable = {} -- diminishing returns table
local targetIcons = {} -- references of target icons
local focusIcons = {} -- references of focus icons
local comboMap = {} -- combo point spell map
local DR_MULTIPLIER = { 1, .5, .25 } -- DR reduction: 1 = 100%, 2 = 50%, 3 = 75%, 4+ = Immune
local NUM_GLYPH_SLOTS = 6

----------------------
--On start
function KnowItAll:OnInitialize()
	self.db = LibStub("AceDB-3.0"):New("KnowItAllDB", defaults, true)
	db = self.db.profile --local the profile table
	self:PopulateOptions(db, media)
    self:RegisterChatCommand(L["knowitall"], "ChatCommand")
    self:RegisterChatCommand(L["kia"], "ChatCommand")
	--callbacks for profile changes
	--self.db.RegisterCallback(self, "OnProfileChanged", "UpdateFrame")
	--self.db.RegisterCallback(self, "OnProfileCopied", "UpdateFrame")
	--self.db.RegisterCallback(self, "OnProfileReset", "UpdateFrame")
	-- New LibBars group
	self.barBox = self:NewBarGroup(self.title, db.orientation, db.bLength)
	self.barBox:SetPoint("TOPLEFT", UIParent, "BOTTOMLEFT", db.cbX, db.cbY)
	self.barBox:SetScale(db.cbScale)
	self.barBox:SetTexture(media:Fetch("statusbar", db.cbTexture))
	self.barBox:SetFont(UNIT_NAME_FONT, 10)
	self.barBox:SetMaxBars(db.maxBars)
	if db.locked then self.barBox:HideAnchor() else self.barBox:ShowAnchor() end
	-- Preallocate all of our target icons
	for i=1,10 do
		local t = self.barBox:CreateTexture(nil, "OVERLAY")
		t:SetTexture("Interface\\AddOns\\KnowItAll\\media\\Target.tga")
		t:SetHeight(15)
		t:SetWidth(15)
		table.insert(targetIcons, t)
	end
	for i=1,10 do
		local t = self.barBox:CreateTexture(nil, "OVERLAY")
		t:SetTexture("Interface\\AddOns\\KnowItAll\\media\\Focus.tga")
		t:SetHeight(15)
		t:SetWidth(15)
		table.insert(focusIcons, t)
	end
	-- Build a combo point spell map
	for k in pairs(self.DebuffTimeId) do
		if type(self.DebuffTimeId[k]) == "table" and type(self.DebuffTimeId[k][2]) == "number" and not (self.DebuffTimeId[k]["talent"] or self.DebuffTimeId[k]["glyph"]) then
			comboMap[k] = true
		end
	end
	for k in pairs(self.CCTimeId) do
		if type(self.CCTimeId[k][2]) == "number" and not (self.CCTimeId[k]["talent"] or self.CCTimeId[k]["glyph"]) then
			comboMap[k] = true
		end
	end
end

----------------------
--Disabled
function KnowItAll:OnDisable()
	if self.frame then
		self.frame:Hide()
	end
end

----------------------
--Loaded
function KnowItAll:OnEnable()
	myGUID = UnitGUID("player")
	_, myClass = UnitClass("player")
	if db.debug then
		self:Print(myClass .. " - " .. myGUID)
		for i = 1, NUM_GLYPH_SLOTS do
			local enabled, glyphType, glyphSpellID = GetGlyphSocketInfo(i)
			if (glyphSpellID) then
				local link = GetGlyphLink(i)
				self:Print(link .. " - " .. glyphSpellID)
			end
		end
	end
	-- Register LibBar callbacks
	self.barBox.RegisterCallback(self, "TimerFinished")
    self.barBox.RegisterCallback(self, "AnchorMoved")
	-- Register WoW events
	self:RegisterEvent("PLAYER_DEAD")
	self:RegisterEvent("PLAYER_ENTERING_WORLD")
	self:RegisterEvent("PLAYER_TARGET_CHANGED")
	self:RegisterEvent("PLAYER_FOCUS_CHANGED")
--	self:RegisterEvent("RAID_TARGET_UPDATE")
	self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
end

----------------------
--Command Handler
function KnowItAll:ChatCommand(input)
    if not input or input:trim() == "" then
        InterfaceOptionsFrame_OpenToCategory(self.optionsFrame)
    else
        LibStub("AceConfigCmd-3.0"):HandleCommand("kia", "KnowItAll", input)
    end
end

----------------------
--Player Died
function KnowItAll:PLAYER_DEAD(event)
	ksCounter = 0 -- reset killing spree counter
end

----------------------
--Player Zoned
function KnowItAll:PLAYER_ENTERING_WORLD(event)
	local inInstance, instanceType = IsInInstance()
	ksCounter = 0 -- reset killing spree counter
	if instanceType == "party" or instanceType == "raid" then
		KnowItAll.inInstance = true
	else
		KnowItAll.inInstance = false
	end
	if instanceType == "arena" or instanceType == "pvp" then
		KnowItAll.inBF = true
	else
		KnowItAll.inBF = false
	end
	if db.debug then
		self:Print("Zoned Into Instance: " .. tostring(KnowItAll.inInstance))
		self:Print("Zoned Into Battlefield: " .. tostring(KnowItAll.inBF))
	end
end

----------------------
--Player Changed Target
function KnowItAll:PLAYER_TARGET_CHANGED(event)
	if not self.barBox:GetBars() then return end
	local count = 1
	for i=1,10 do
		targetIcons[i]:Hide()
		targetIcons[i]:ClearAllPoints()
	end
	for k, v in pairs(self.barBox:GetBars()) do
		-- BUG: for some reason our tables don't always match up.  not sure why
		if barTable[tonumber(string.sub(k, 2))] and barTable[tonumber(string.sub(k, 2))][2] == UnitGUID("target") and count < 11 then
			targetIcons[count]:SetParent(v)
			if db.orientation % 2 == 0 then
	   			targetIcons[count]:SetPoint("TOPLEFT", v, "BOTTOMLEFT", 0, -15)
			else
				targetIcons[count]:SetPoint("BOTTOMRIGHT", v, "BOTTOMLEFT", -15, 0)
			end
			targetIcons[count]:Show()
			count = count + 1
		end
	end
end

----------------------
--Player Changed Focus
function KnowItAll:PLAYER_FOCUS_CHANGED(event)
	if not self.barBox:GetBars() then return end
	local count = 1
	for i=1,10 do
		focusIcons[i]:Hide()
		focusIcons[i]:ClearAllPoints()
	end
	for k, v in pairs(self.barBox:GetBars()) do
		-- BUG: for some reason our tables don't always match up.  not sure why
		if barTable[tonumber(string.sub(k, 2))] and barTable[tonumber(string.sub(k, 2))][2] == UnitGUID("focus") and count < 11 then
			focusIcons[count]:SetParent(v)
			if db.orientation % 2 == 0 then
				focusIcons[count]:SetPoint("TOPLEFT", v, "BOTTOMLEFT", 0, -15 * 2)
			else
				focusIcons[count]:SetPoint("BOTTOMRIGHT", v, "BOTTOMLEFT", -15 * 2, 0)
			end
			focusIcons[count]:Show()
			count = count + 1
		end
	end
end

----------------------
--Raid Icons Changed
function KnowItAll:RAID_TARGET_UPDATE(event)
	self:Print(GetRaidTargetIndex("target"))
end

----------------------
--Combat events
function KnowItAll:COMBAT_LOG_EVENT_UNFILTERED(event, timestamp, ce, sGUID, sName, sFlags, dGUID, dName, dFlags, spellId, ...)
	if db.debug and spellId then
		local spellName,spellRank = GetSpellInfo(spellId)
		--self:Print(ce .. "::" .. sGUID .. "::" .. dGUID)
		--self:Print(ce .. "::" .. sFlags .. "::" .. dFlags)
		self:Print(ce .. "::" .. tostring(bit.band(sFlags, 0x0000000F)) .. "::" .. tostring(bit.band(dFlags, 0x0000000F)))
		self:Print(tostring(spellName) .. ' (' .. tostring(spellRank) .. ') - ' .. spellId)
		--self:Print(GetComboPoints("player"))
	end

	--check for events we care about
	if not COMBAT_EVENTS[ce] then return end

	local spellRank, spellName, spellTexture, spellTime, spellColor
	local _, _, delay = GetNetStats()
	delay = delay / 1000
	------------Track Summons----------------
	if ce == "SPELL_SUMMON" then
		if not db.summonBars or
		  (db.summonHideI and self.inInstance) or
		  (db.summonHideB and self.inBF) or
		  not self.SummonTimeId[spellId] then return end
		if bit.band(sFlags, 0xF) == 0x1 then
			dName = UnitName("player")
		else
		    dName = UnitName("none")
		end
		spellTime = self:getTime(self.SummonTimeId[spellId], sGUID)
	    spellColor = db.summonColor
	------------Track HoTs, DoTs, CCs and invulns----------------
	elseif ce == "SPELL_AURA_APPLIED" then
		-- combo point based spell? we do this in SPELL_CAST_SUCCESS
		if comboMap[spellId] then return end
		-- some spells like winter's chill are debuffs with no source id
		-- but we want to track them anyway so we need this special case
		if not sGUID and db.debuffBars and db.globalDebuffs and self.globalSpells[myClass][spellId] then
			if (db.debuffHideI and self.inInstance) or (db.debuffHideB and self.inBF) then return end
			spellTime = self.DebuffTimeId[spellId]
			spellColor = db.debuffColor
		end
		-- we only want to track HoTs/DoTs we cast but include global spells and exclude the ones we cast on ourselves
		if (sGUID == myGUID or (self.globalSpells[myClass][spellId] and bit.band(sFlags, 0xF0) == 0x10)) and dGUID ~= myGUID then
			if db.debuffBars and self.DebuffTimeId[spellId] then
				if (db.debuffHideI and self.inInstance) or (db.debuffHideB and self.inBF) then return end
				if sGUID ~= myGUID and not db.globalDebuffs then return end
				spellTime = self:getTime(self.DebuffTimeId[spellId], sGUID, self:IsHuman(dGUID))
				spellColor = db.debuffColor
			elseif db.gainBars and self.BuffTimeId[spellId] then
				if (db.gainHideI and self.inInstance) or (db.gainHideB and self.inBF) then return end
				if sGUID ~= myGUID and not db.globalGains then return end
				spellTime = self:getTime(self.BuffTimeId[spellId], sGUID)
				spellColor = db.gainColor
			end
		end
		-- we want to track everyone's CC and invuln effects
		if db.ccBars and self.CCTimeId[spellId] then
			if (db.ccHideI and self.inInstance) or (db.ccHideB and self.inBF) then return end
			spellTime = self:getTime(self.CCTimeId[spellId], sGUID)
			-- if our target is human we have a cap on our CCs
			if self:IsHuman(dGUID) and spellTime > 10 then spellTime = self.CCTimeId[spellId].pvp or 10 end
			-- check if we need DRs
			if self.CCTimeId[spellId]["cat"] and (self:IsHuman(dGUID) or self.CCTimeId[spellId]["pve"]) then
				spellTime = KnowItAll:SetDR(dGUID, self.CCTimeId[spellId]["cat"], spellTime)
			end
			spellColor = db.ccColor
		elseif db.invulnBars and self.InvulnTimeId[spellId] then
			if (db.invulnHideI and self.inInstance) or (db.invulnHideB and self.inBF) then return end
		    spellTime = self:getTime(self.InvulnTimeId[spellId], sGUID)
		    spellColor = db.invulnColor
		end
	------------Update our spell stacks and aura refreshes----------------
	elseif ce == "SPELL_AURA_APPLIED_DOSE" or ce == "SPELL_AURA_REFRESH" or ce == "SPELL_AURA_REMOVED_DOSE" then
		local stack = select(4, ...)
		if ce ~= "SPELL_AURA_REMOVED_DOSE" then
			-- look for a matching bar
			for k, v in pairs(barTable) do
				local a, b, c = unpack(v)
				if (a == sGUID and b == dGUID and c == spellId) then
					-- Restart the timer
					local refresh = self.barBox:GetBar("B" .. k)
					local rtime = refresh.maxValue
					refresh:StopTimer()
					-- Check for DR
					if self.CCTimeId[spellId] and self.CCTimeId[spellId]["cat"] and (self:IsHuman(dGUID) or self.CCTimeId[spellId]["pve"]) then
						rtime = KnowItAll:SetDR(dGUID, self.CCTimeId[spellId]["cat"], refresh.maxValue)
					end
					refresh:SetTimer(rtime - delay, refresh.maxValue)
				end
			end
		end
	    if stack and stack > 0 then
   			--iconText = icon:CreateFontString(nil, "OVERLAY", "NumberFontNormalSmall")
			--iconText:SetPoint("BOTTOMRIGHT", icon, 5, 0)
			--iconText:SetFont("ARIALN.TTF",12,"MONOCHROME")
			--iconText:SetText(stack)
	    end
	------------Track all spell aura loses----------------
	elseif ce == "SPELL_AURA_REMOVED" then
		self:DeleteBar(dGUID, spellId)
	------------Combo point spells----------------
	elseif ce == "SPELL_CAST_SUCCESS" then
		-- combo points are only correct when SPELL_CAST_SUCCESS fires so we need this here
		if not comboMap[spellId] then return end
		local comboPoints = 1
		if sGUID == myGUID then
			comboPoints = GetComboPoints("player")
		elseif sGUID == UnitGUID("target") then
			comboPoints = GetComboPoints("target")  -- no idea if this works
		end
		if comboPoints < 1 then comboPoints = 1 end
		if db.ccBars and self.CCTimeId[spellId] then
			if (db.ccHideI and self.inInstance) or (db.ccHideB and self.inBF) then return end
			spellTime = self.CCTimeId[spellId][comboPoints]
			spellColor = db.ccColor
		elseif (sGUID == myGUID or (self.globalSpells[myClass][spellId] and bit.band(sFlags, 0xF0) == 0x10)) and db.debuffBars and self.DebuffTimeId[spellId] then
			if (db.debuffHideI and self.inInstance) or (db.debuffHideB and self.inBF) then return end
			if sGUID ~= myGUID and not db.globalDebuffs then return end
			spellTime = self.DebuffTimeId[spellId][comboPoints]
			spellColor = db.debuffColor
		end
	------------deaths----------------
	elseif ce == "PARTY_KILL" then
		-- All hail UT2k4
		if db.sounds and sGUID == myGUID and self:IsHuman(dGUID) then
			ksCounter = ksCounter + 1
			if ksCounter == 2 then
				PlaySoundFile(media:Fetch("sound", "UT2K4: Spree"));
			elseif ksCounter == 3 then
				PlaySoundFile(media:Fetch("sound", "UT2K4: Rampage"));
			elseif ksCounter == 4 then
				PlaySoundFile(media:Fetch("sound", "UT2K4: Dominating"));
			elseif ksCounter == 5 then
				PlaySoundFile(media:Fetch("sound", "UT2K4: Unstoppable"));
			elseif ksCounter == 6 then
				PlaySoundFile(media:Fetch("sound", "UT2K4: Godlike"));
			elseif ksCounter > 6 then
				PlaySoundFile(media:Fetch("sound", "UT2K4: Wicked"));
			end
		end
	elseif ce == "UNIT_DIED" or ce == "UNIT_DESTROYED" then
		-- Someone died, delete all their bars
		self:DeleteBar(dGUID)
	end

	if spellId and spellTime and spellTime > 0 and dName then
		local spellName,spellRank,spellTexture = GetSpellInfo(spellId)
		local flash = 0
		local r,g,b = unpack(spellColor)
		local bName = self:BarIdGen(sGUID,dGUID,spellId)
		if db.flash and spellTime > 8 and self.CCTimeId[spellId] then flash = 5 end  --flash if long CC spell
		if rawget(L,spellName) then spellName = L[spellName] end  -- check for localized abbr
		if bit.band(dFlags, 0xFC00) == 0x400 then dName = string.match(dName, "[^-]*") end  -- strip realm off of players
		if sGUID ~= myGUID and self.globalSpells[myClass][spellId] then spellName = "[" .. spellName .. "]" end
		--self:Print("New Bar: " .. spellName .. " for " .. (spellTime - delay) .. " sec.")
		local ntb = self.barBox:NewTimerBar(bName, spellName .. " - " .. dName, spellTime - delay, spellTime, spellTexture, flash)
		ntb:SetColorAt(0, r, g, b, 1)
		if UnitGUID("target") == dGUID then self:PLAYER_TARGET_CHANGED() end
		if UnitGUID("focus") == dGUID then self:PLAYER_FOCUS_CHANGED() end
	end
end

----------------------
--Calc time for a spell
function KnowItAll:getTime(entry, sGUID, isPvP)
	if type(entry) == "table" then
		-- Check if spell is affected by talents
		if entry["talent"] and sGUID == myGUID then
			return entry[select(5, GetTalentInfo(entry["talent"][1], entry["talent"][2]))+1]
		-- or by glyphs
		elseif entry["glyph"] and sGUID == myGUID then
			for i = 1, NUM_GLYPH_SLOTS do
				local _, _, glyphSpellID = GetGlyphSocketInfo(i)
				if glyphSpellID == entry["glyph"] then
					return entry[2]
				end
			end
		end
		-- else return special pvp time or return the default first value
		if isPvP and entry.pvp then return entry.pvp else return entry[1] end
	else
		return entry
	end
end

----------------------
--Is the GUID a human player (our CCs can only be 10s max)
function KnowItAll:IsHuman(GUID)
	if strsub(GUID, 1, 5) == "0x000" then
		return true
	else
		return false
	end
end

----------------------
--Bar ID Generator
function KnowItAll:BarIdGen(sGUID,dGUID,spellId)
	-- restart our old bar if we cast the same spell
	for k, v in pairs(barTable) do
		local a, b, c = unpack(v)
		if (a == sGUID and b == dGUID and c == spellId) then
			--self:Print("Bar Found! Restarting " .. k)
			return "B" .. k
		end
	end
	-- TODO: table recycle
	if #barTable > 0 then
		self.barNum = self.barNum + 1
	else
		self.barNum = 1 -- reset bar ID counter
	end
	barTable[self.barNum] = {sGUID,dGUID,spellId,1}
	return "B" .. self.barNum
end

----------------------
--Bar Deleter
function KnowItAll:DeleteBar(dGUID, spellId)
	for k, v in pairs(barTable) do
		local _, a, b = unpack(v)
		if (a == dGUID and (b == spellId or spellId == nil)) then
			--self:Print("Bar Found! Deleting " .. k)
			self.barBox:RemoveBar("B" .. k)
			barTable[k] = nil
		end
	end
	-- Reschedule DRs if needed
	if drTable[dGUID] and self.CCTimeId[spellId] and self.CCTimeId[spellId].cat and drTable[dGUID][self.CCTimeId[spellId].cat] then
		if drTable[dGUID][self.CCTimeId[spellId].cat].handle and self:TimeLeft(drTable[dGUID][self.CCTimeId[spellId].cat].handle) > 15 then
			self:CancelTimer(drTable[dGUID][self.CCTimeId[spellId].cat].handle, true)
			local handle = self:ScheduleTimer("RemoveDR", 15, {dGUID, self.CCTimeId[spellId].cat})
			drTable[dGUID][self.CCTimeId[spellId].cat].handle = handle
		end
	end
end

----------------------
--Set DRs
function KnowItAll:SetDR(GUID, cat, spelltime)
	local stack = 0
	drTable[GUID] = drTable[GUID] or {}
	if drTable[GUID][cat] then
		self:CancelTimer(drTable[GUID][cat].handle, true)
		stack =	drTable[GUID][cat].stack
	end
	if stack < 3 then
		stack =	stack + 1
		spelltime = spelltime * DR_MULTIPLIER[stack]
		local handle = self:ScheduleTimer("RemoveDR", 15 + spelltime, {GUID, cat})
		drTable[GUID][cat] = { ["stack"] = stack, ["handle"] = handle }
		return spelltime
	end
	return 0
end

----------------------
--Remove active DR
function KnowItAll:RemoveDR(args) -- [1] = GUID, [2] = cat
	local GUID, cat = args[1], args[2]
	if drTable[GUID] then
		if drTable[GUID][cat] then
			drTable[GUID][cat] = nil
		end
		if table.getn(drTable[GUID]) == 0 then
		    drTable[GUID] = nil
		end
	end
end

-- LibBar Call Backs
function KnowItAll:AnchorMoved(_, _, x, y)
	if db.debug then self:Print(tostring(x) .. "," .. tostring(y)) end
    db.cbX, db.cbY = x, y
end

function KnowItAll:TimerFinished(_, group, bar, name)
	if string.sub(name, 1, 1) == "B" then
		barTable[tonumber(string.sub(name, 2))] = nil
	end
end
