local KnowItAll = KnowItAll
local L = LibStub("AceLocale-3.0"):GetLocale("KnowItAll", true)
local db

function KnowItAll:PopulateOptions(mydb, media)
	db = mydb
	local options = {
		name = "KnowItAll",
		type = "group",
		childGroups = "tab",
		args = {
			bars = {
				type="group",
				name=L["Bars"],
				order = 1,
				args = {
					count = {
						type = "range",
						name = L["Max Bar Count"],
						min = 1,
						max = 20,
						step = 1,
						order = 1,
				        get = function(info) return db.maxBars end,
				        set = function(info,val) self.barBox:SetMaxBars(val); db.maxBars = val end
					},
					length = {
						type = "range",
						name = L["Bar Length"],
						min = 100,
						max = 300,
						step = 10,
						order = 2,
				        get = function(info) return db.bLength end,
				        set = function(info,val) self.barBox:SetLength(val); db.bLength = val end
					},
					scale = {
						type = "range",
						name = L["Bar Scale"],
						min = 0.5,
						max = 1.5,
						step = 0.1,
						order = 3,
				        get = function(info) return db.cbScale end,
				        set = function(info,val) self.barBox:SetScale(val); db.cbScale = val end
					},
					spacing = {
						type = "range",
						name = L["Bar Spacing"],
						min = 0,
						max = 10,
						step = 1,
						order = 4,
				        get = function(info) return db.cbSpacing end,
				        set = function(info,val) self.barBox:SetSpacing(val); db.cbSpacing = val end
					},
					orientation = {
						type = "select",
						name = L["Bar Orientation"],
						desc = L["Bar growth direction."],
						order = 5,
						values = {L["Left-to-Right"], L["Bottom-to-Top"], L["Right-to-Left"], L["Top-to-Bottom"]},
						get = function(info) return db.orientation end,
						set = function(info,val)
							db.orientation = val
							self.barBox:SetOrientation(val)
						end
					},
					texture = {
						type = "select",
						name = L["Texture"],
						desc = L["Texture"],
						order = 6,
						values = media:List("statusbar"),
						get = function(info)
							for k,v in pairs(media:List("statusbar")) do
								if v == db.cbTexture then return k end
							end
							return 0
						end,
						set = function(info,val)
							db.cbTexture = media:List("statusbar")[val]
							self.barBox:SetTexture(media:Fetch("statusbar", db.cbTexture))
						end
					},
				    lock = {
				        type = "toggle",
				        name = L["Lock"],
				        desc = L["Locks the bar frame."],
				        order = 7,
				        get = function(info) return not self.barBox:IsAnchorVisible() end,
				        set = function(info,val) if val then self.barBox:HideAnchor(); db.locked = true else self.barBox:ShowAnchor(); db.locked = false end end
				    },
				    flash = {
				        type = "toggle",
				        name = L["CC Bar Flash"],
				        desc = L["On long CCs the bar will flash 5 seconds before the end."],
				        order = 8,
				        get = function(info) return db.flash end,
				        set = function(info,val) db.flash = val end
				    },
					bartest = {
						type="execute",
				        name=L["Bar Test"],
				        desc=L["Display a set of test bars."],
				        func = function() KnowItAll:BarTest() end,
					}
				}
			},
			spells = {
				type="group",
				name=L["Spells"],
				order = 2,
				args = {
					ggrp = {
						type="group",
						name=L["Gains"],
						order = 1,
						args = {
							gains = {
							    type = "toggle",
							    name = L["Enable"],
							    desc = L["Enable gain bars."],
							    order = 1,
							    get = function(info) return db.gainBars end,
							    set = function(info,val) db.gainBars = val end
							},
							hidei = {
							    type = "toggle",
							    name = L["Hide in instances"],
							    desc = L["Don't show this bar type when inside instances."],
							    order = 2,
							    get = function(info) return db.gainHideI end,
							    set = function(info,val) db.gainHideI = val end
							},
							hidebf = {
							    type = "toggle",
							    name = L["Hide in battlefields"],
							    desc = L["Don't show this bar type when inside battlegrounds or arenas."],
							    order = 3,
							    get = function(info) return db.gainHideB end,
							    set = function(info,val) db.gainHideB = val end
							},
							global = {
							    type = "toggle",
							    name = L["Shared Class Spells"],
							    desc = L["Track other friendly player's spells of the same class that cannot be applied to a target more than once. (Sunder armor, fear ward, etc.)"],
							    order = 4,
							    get = function(info) return db.globalGains end,
							    set = function(info,val) db.globalGains = val end
							},
							gainscolor = {
								type = "color",
								name = L["Bar Color"],
								desc = L["Change this spell type's bar color."],
								get = function(info)
									local r, g, b = unpack(db.gainColor)
									return r, g, b
								end,
								set = function(info, r, g, b) db.gainColor = { r, g, b } end
							}
						}
					},
					dgrp = {
						type="group",
						name=L["Debuffs"],
						order = 2,
						args = {
						    debuffs = {
						        type = "toggle",
						        name = L["Enable"],
						        desc = L["Enable debuff bars."],
							    order = 1,
						        get = function(info) return db.debuffBars end,
						        set = function(info,val) db.debuffBars = val end
						    },
							hidei = {
							    type = "toggle",
							    name = L["Hide in instances"],
							    desc = L["Don't show this bar type when inside instances."],
							    order = 2,
							    get = function(info) return db.debuffHideI end,
							    set = function(info,val) db.debuffHideI = val end
							},
							hidebf = {
							    type = "toggle",
							    name = L["Hide in battlefields"],
							    desc = L["Don't show this bar type when inside battlegrounds or arenas."],
							    order = 3,
							    get = function(info) return db.debuffHideB end,
							    set = function(info,val) db.debuffHideB = val end
							},
							global = {
							    type = "toggle",
							    name = L["Shared Class Spells"],
							    desc = L["Track other friendly player's spells of the same class that cannot be applied to a target more than once. (Sunder armor, fear ward, etc.)"],
							    order = 4,
							    get = function(info) return db.globalDebuffs end,
							    set = function(info,val) db.globalDebuffs = val end
							},
							debuffscolor = {
								type = "color",
								name = L["Bar Color"],
								desc = L["Change this spell type's bar color."],
								get = function(info)
									local r, g, b = unpack(db.debuffColor)
									return r, g, b
								end,
								set = function(info, r, g, b) db.debuffColor = { r, g, b } end
							}
						}
					},
					ccgrp = {
						type="group",
						name=L["CCs"],
						order = 3,
						args = {
						    cc = {
						        type = "toggle",
						        name = L["Enable"],
						        desc = L["Enable tracking of crowd control effects."],
							    order = 1,
						        get = function(info) return db.ccBars end,
						        set = function(info,val) ccBars = val end
						    },
							hidei = {
							    type = "toggle",
							    name = L["Hide in instances"],
							    desc = L["Don't show this bar type when inside instances."],
							    order = 2,
							    get = function(info) return db.ccHideI end,
							    set = function(info,val) db.ccHideI = val end
							},
							hidebf = {
							    type = "toggle",
							    name = L["Hide in battlefields"],
							    desc = L["Don't show this bar type when inside battlegrounds or arenas."],
							    order = 3,
							    get = function(info) return db.ccHideB end,
							    set = function(info,val) db.ccHideB = val end
							},
							cccolor = {
								type = "color",
								name = L["Bar Color"],
								desc = L["Change this spell type's bar color."],
								get = function(info)
									local r, g, b = unpack(db.ccColor)
									return r, g, b
								end,
								set = function(info, r, g, b) db.ccColor = { r, g, b } end
							}
						}
					},
					invulngrp = {
						type="group",
						name=L["Invulnerabilities"],
						order = 4,
						args = {
						    invuln = {
						        type = "toggle",
						        name = L["Enable"],
						        desc = L["Enable tracking of invulnerability/absorption effects."],
						        order = 1,
						        get = function(info) return db.invulnBars end,
						        set = function(info,val) db.invulnBars = val end
						    },
							hidei = {
							    type = "toggle",
							    name = L["Hide in instances"],
							    desc = L["Don't show this bar type when inside instances."],
							    order = 2,
							    get = function(info) return db.invulnHideI end,
							    set = function(info,val) db.invulnHideI = val end
							},
							hidebf = {
							    type = "toggle",
							    name = L["Hide in battlefields"],
							    desc = L["Don't show this bar type when inside battlegrounds or arenas."],
							    order = 3,
							    get = function(info) return db.invulnHideB end,
							    set = function(info,val) db.invulnHideB = val end
							},
							invulncolor = {
								type = "color",
								name = L["Bar Color"],
								desc = L["Change this spell type's bar color."],
								get = function(info)
									local r, g, b = unpack(db.invulnColor)
									return r, g, b
								end,
								set = function(info, r, g, b) db.invulnColor = { r, g, b } end
							}
						}
					},
					sumgrp = {
						type="group",
						name=L["Summons"],
						order = 5,
						args = {
						    summons = {
						        type = "toggle",
						        name = L["Enable"],
						        desc = L["Enable tracking of temporarily summoned units."],
						        order = 1,
						        get = function(info) return db.summonBars end,
						        set = function(info,val) db.summonBars = val end
						    },
							hidei = {
							    type = "toggle",
							    name = L["Hide in instances"],
							    desc = L["Don't show this bar type when inside instances."],
							    order = 2,
							    get = function(info) return db.summonHideI end,
							    set = function(info,val) db.summonHideI = val end
							},
							hidebf = {
							    type = "toggle",
							    name = L["Hide in battlefields"],
							    desc = L["Don't show this bar type when inside battlegrounds or arenas."],
							    order = 3,
							    get = function(info) return db.summonHideB end,
							    set = function(info,val) db.summonHideB = val end
							},
							sumcolor = {
								type = "color",
								name = L["Bar Color"],
								desc = L["Change this spell type's bar color."],
								get = function(info)
									local r, g, b = unpack(db.summonColor)
									return r, g, b
								end,
								set = function(info, r, g, b) db.summonColor = { r, g, b } end
							}
						}
					},
					globalgrp = {
						type="group",
						name=L["Globals"],
						order = 6,
						args = {
						}
					}
				}
			},
			debug = {
				type="group",
				name=L["Misc"],
				order = 3,
				args = {
				    version = {
				        type = "description",
				        name = KnowItAll.title .. "\nv" .. tostring(KnowItAll.version) .. "\n\n",
						order = 1,
				    },
				    sounds = {
				        type = "toggle",
				        name = L["Sounds"],
				        desc = L["Turns killing spree sounds on or off."],
				        order = 2,
				        get = function(info) return db.sounds end,
				        set = function(info,val) db.sounds = val end
				    },
				    debug = {
				        type = "toggle",
				        name = L["Debug"],
				        desc = L["Turns debugging output on or off."],
						order = 3,
				        get = function(info) return db.debug end,
				        set = function(info,val) db.debug = val end
				    }
				}
			}
		}
	}
    self.optionsFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("KnowItAll", self.title)
	LibStub("AceConfig-3.0"):RegisterOptionsTable("KnowItAll", options)
end

----------------------
--Bar Test
function KnowItAll:BarTest()
	for i=1,db.maxBars do
		local tb = self.barBox:NewTimerBar("T" .. i, "Test Bar " .. i, i+5, nil, nil, 0)
	end
end
